// General-purpose Photovoltaic Device Model gpvdm.com - a drift diffusion
// base/Shockley-Read-Hall model for 1st, 2nd and 3rd generation solarcells.
// The model can simulate OLEDs, Perovskite cells, and OFETs.
// 
// Copyright (C) 2008-2020 Roderick C. I. MacKenzie
// All rights reserved.
// 
// https://www.gpvdm.com
// r.c.i.mackenzie at googlemail.com
// 
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//     * Neither the name of the GPVDM nor the
//       names of its contributors may be used to endorse or promote products
//       derived from this software without specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
// DISCLAIMED. IN NO EVENT SHALL Roderick C. I. MacKenzie BE LIABLE FOR ANY
// DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
// ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "../../sim.h"
#include "../../dump.h"
#include "tpv.h"
#include "wire.h"
#include "../../dynamic_store.h"

static int unused __attribute__((unused));

struct tpv tpv_config;

void sim_tpv(struct device *in)
{
tpv_load_config(&tpv_config,in);

int tpv_running=FALSE;
in->dt=tpv_config.tpvdtstart;


carrier_count_reset(in);

in->go_time=FALSE;

struct istruct tpvvoc;
in->Vapplied=tpv_config.tpvvoc;
inter_init(&tpvvoc);

struct dynamic_store store;
dump_dynamic_init(&store,in);

int step=0;




if (tpv_config.tpv_mode==tpv_set_voltage)
{
printf("Running tpv %e V\n",tpv_config.tpvvoc);
in->Psun=1.0;
light_solve_and_update(in,&(in->mylight),in->Psun,0.0,0.0);
set_light_for_voc(in,tpv_config.tpvvoc);
in->Vapplied=tpv_config.tpvvoc;
}else
{
printf("Running tpv for a light intensity of %e W\n",in->Psun);
light_solve_and_update(in,&(in->mylight),in->Psun,0.0,0.0);
newton_sim_tpv(in);
}

newton_sim_tpv_fast(in,FALSE);
in->ilast=get_I(in)+in->Vapplied/in->Rshunt;
in->Vapplied_last=in->Vapplied;

time_init(in);
device_timestep(in);
in->dt=tpv_config.tpvdtstart;
in->go_time=TRUE;
in->time=tpv_config.tpvstart;

newton_sim_tpv_fast(in,TRUE);


FILE *out=fopen("./tpvpsun.dat","w");
fprintf(out,"%le\n",in->Psun);
fclose(out);

FILE *tpvn=fopen("./tpv_cden.dat","w");
fprintf(tpvn,"%le\n",get_extracted_np(in));
fclose(tpvn);


do
{
	if (in->time<0.0)
	{
		light_solve_and_update(in,&(in->mylight),in->Psun,0.0,0.0);
		reset_np_save(in);
	}else
	if ((in->time>=0.0)&&(in->time<(tpv_config.tpv_laser_width)))
	{
		if (tpv_config.tpv_excitation_mode==tpv_mode_laser)
		{
			light_solve_and_update(in,&(in->mylight),in->Psun,tpv_config.tpvlaserpower,tpv_config.tpv_laser_width);
			int ii;

			if (get_dump_status(dump_iodump)==TRUE)
			{
				FILE *out =fopena(in->outputpath,"./Gn_TPV_laser.dat","w");
				for (ii=0;ii<in->ymeshpoints;ii++)
				{
						fprintf(out,"%e %le\n",in->ymesh[ii],in->Gn[ii]);
				}
				fclose(out);
			}
		}
	}else
	{
		if (tpv_running==FALSE)
		{
			tpv_running=TRUE;
		}

		if (tpv_config.tpv_excitation_mode==tpv_mode_laser)
		{
			light_solve_and_update(in,&(in->mylight),in->Psun,0.0,0.0);
		}else
		if (tpv_config.tpv_excitation_mode==tpv_mode_sun)
		{
			light_solve_and_update(in,&(in->mylight),0.0,0.0,0.0);
		}
	}


	newton_sim_tpv_fast(in,TRUE);

	if (get_dump_status(dump_print_converge)==TRUE)
	{
	printf("tpv time=%e step=%d %.1e ",in->time,step,in->last_error);
	printf("V=%lf current = %e mA (%e A/m^2)\n",in->Vapplied,get_I(in)/1e-3,get_J(in));

	}


	if (tpv_running==TRUE)
	{

		dump_dynamic_add_data(&store,in,in->time);
	}

	inter_append(&tpvvoc,in->time,get_V_TPV(in));

	plot_now(in,"tpv_vars.plot");

	if (in->time>tpv_config.tpv_laser_width) in->dt*=tpv_config.tpvdtmull;

	if (step==tpv_config.tpvpulsestart)
	{
		in->dt=tpv_config.tpvdtstart;
	}

	step++;
	device_timestep(in);



	if (in->time>tpv_config.tpvstop) break;

	if (in->stop==TRUE)
	{
		break;

	}

in->ilast=get_I(in)+in->C*(in->Vapplied-in->Vapplied_last)/in->dt+in->Vapplied/in->Rshunt;

}while(1);

in->go_time=FALSE;

char outpath[200];



sprintf(outpath,"%s%s",in->outputpath,"tpv_voc.dat");
inter_save(&tpvvoc,outpath);

inter_sub_double(&tpvvoc,tpvvoc.data[0]);
sprintf(outpath,"%s%s",in->outputpath,"tpv_voc_norm.dat");
inter_save(&tpvvoc,outpath);

inter_free(&tpvvoc);

dump_dynamic_save(in->outputpath,&store);
dump_dynamic_free(&store);

}

void tpv_load_config(struct tpv *in,struct device *dev)
{
double ver=0.0;
char name[200];
FILE *config=fopena(dev->outputpath,"tpv.inp","r");

config_read_line_to_double(&(in->tpvdtstart),config,"#tpvdtstart");

config_read_line_to_double(&(in->tpv_laser_width),config,"#tpv_laser_width");

config_read_line_to_double(&(in->tpvlaserpower),config,"#tpvlaserpower");
in->tpvlaserpower=fabs(in->tpvlaserpower);

config_read_line_to_double(&(in->tpvdtmull),config,"#tpvdtmull");

config_read_line_to_double(&(in->tpvstop),config,"#tpvstop");

config_read_line_to_double(&(in->tpvvoc),config,"#tpvvoc");

config_read_line_string_decode_to_int(&(in->tpv_mode),config,"#tpv_mode");

config_read_line_string_decode_to_int(&(in->tpv_excitation_mode),config,"#tpv_excitation_mode");

config_read_line_to_double(&(in->Rscope),config,"#Rscope");

config_read_line_to_double(&(in->tpv_L),config,"#tpv_L");

config_read_line_to_int(&(in->simwire),config,"#sim_wire");

unused=fscanf(config,"%s",name);
unused=fscanf(config,"%lf",&(ver));
if (ver!=1.1)
{
ewe("File compatability problem %s\n","tpv.inp");
}

unused=fscanf(config,"%s",name);
if (strcmp("#end",name)!=0)
{
ewe("Problem with tpv.inp file last item read %s!\n",name);
}
fclose(config);


}

double get_V_TPV(struct device *in)
{
double ret=tpv_config.Rscope*in->Vapplied/(in->Rcontact+tpv_config.Rscope);

return ret;
}

